@@ -66,6 +66,7 @@ class AgentClient:
         bottle.setQueryNodes(['P1'])
 
         """
+        self.session = requests.Session()
         self.genome = None
         self._bottle_info = bottle_info
         self.name = bottle_info['name']
@@ -82,10 +83,16 @@ class AgentClient:
         self.summarize_for_single_node = True
         if 'secure' not in self._bottle_info or self._bottle_info['secure']:
             self._secure = True
-            self._url = 'https://{name}.{domain}/'.format(**self._bottle_info)
+            if not self.name:
+                self._url = 'https://{domain}/'.format(**self._bottle_info)
+            else:
+                self._url = 'https://{name}.{domain}/'.format(**self._bottle_info)
         else:
             self._secure = False
-            self._url = 'http://{name}.{domain}/'.format(**self._bottle_info)
+            if not self.name:
+                self._url = 'http://{domain}/'.format(**self._bottle_info)
+            else:
+                self._url = 'http://{name}.{domain}/'.format(**self._bottle_info)
 
     def __repr__(self) -> str:
         return (
@@ -107,8 +114,8 @@ class AgentClient:
         return self.send_unique_ids
 
     def connect(self) -> Dict:
-        """Grabs the bottle's gaius_agent's genome for node definitions."""
-        response_data = requests.get(self._url + 'connect', verify=self._secure, headers=self._headers).json()
+        """Establishes initial connection to GAIuS agent and grabs the bottle's gaius_agent's genome for node definitions."""
+        response_data = self.session.get(self._url + 'connect', verify=self._secure, headers=self._headers).json()
         if 'status' not in response_data or response_data['status'] != 'okay':
             self._connected = False
             raise AgentConnectionError("Connection failed!", response_data)
@@ -151,7 +158,9 @@ class AgentClient:
                 data['unique_id'] = unique_id
             else:
                 data = {'unique_id': unique_id}
-
+                
+        data = json.loads(json.dumps(data))
+        
         if isinstance(nodes[0], str):
             nodes = [{'name': name, 'id': self.genome.primitive_map[name]} for name in nodes]
         for node in nodes:
@@ -187,7 +196,7 @@ class AgentClient:
         """Exclusively uses the 'observe' call."""
         if nodes is None:
             nodes = self.ingress_nodes
-        return self._query(requests.post, 'observe', data=data, nodes=nodes)
+        return self._query(self.session.post, 'observe', data=data, nodes=nodes)
 
     def _observe_event(self, data: Dict, unique_id: str = None) -> Tuple[dict, str]:
         """Exclusively uses the 'observe' call."""
@@ -196,7 +205,7 @@ class AgentClient:
         if unique_id is None:
             unique_id = str(uuid.uuid4())
         for node, node_data in data.items():
-            response, uid = self._query(requests.post, 'observe', data=node_data, nodes=[node], unique_id=unique_id)
+            response, uid = self._query(self.session.post, 'observe', data=node_data, nodes=[node], unique_id=unique_id)
             results[node] = response
         return results, uid
 
@@ -208,55 +217,55 @@ class AgentClient:
         """
         if nodes is None:
             nodes = self.query_nodes
-        return self._query(requests.post, 'observe', data=data, nodes=nodes)
+        return self._query(self.session.post, 'observe', data=data, nodes=nodes)
 
     def show_status(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Return the current status of the bottle."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.get, 'status', nodes=nodes)
+        return self._query(self.session.get, 'status', nodes=nodes)
 
     def learn(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Return the learn results."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.post, 'learn', nodes=nodes)
+        return self._query(self.session.post, 'learn', nodes=nodes)
 
     def get_wm(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Return information about Working Memory."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.get, 'working-memory', nodes=nodes)
+        return self._query(self.session.get, 'working-memory', nodes=nodes)
 
     def get_predictions(self, unique_id: str = None, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Return prediction result data."""
         if nodes is None:
             nodes = self.query_nodes
-        return self._query(requests.post, 'predictions', nodes=nodes, unique_id=unique_id)
+        return self._query(self.session.post, 'predictions', nodes=nodes, unique_id=unique_id)
 
     def clear_wm(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Clear the Working Memory of the Genie."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.post, 'working-memory/clear', nodes=nodes)
+        return self._query(self.session.post, 'working-memory/clear', nodes=nodes)
 
     def clear_all_memory(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Clear both the Working Memory and persisted memory."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.post, 'clear-all-memory', nodes=nodes)
+        return self._query(self.session.post, 'clear-all-memory', nodes=nodes)
 
     def get_percept_data(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Return percept data."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.get, 'percept-data', nodes=nodes)
+        return self._query(self.session.get, 'percept-data', nodes=nodes)
 
     def get_cognition_data(self, unique_id: str = None, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Return cognition data."""
         if nodes is None:
             nodes = self.query_nodes
-        return self._query(requests.get, 'cognition-data', nodes=nodes, unique_id=unique_id)
+        return self._query(self.session.get, 'cognition-data', nodes=nodes, unique_id=unique_id)
 
     @_ensure_connected
     def change_genes(self, gene_data: Dict, nodes: List = None) -> Union[dict, Any]:
@@ -276,7 +285,7 @@ class AgentClient:
 
         result = {}
         for node in nodes:
-            response = requests.post(
+            response = self.session.post(
                 f"{self._url}{node['id']}/genes/change",
                 verify=self._secure,
                 headers=self._headers,
@@ -301,7 +310,7 @@ class AgentClient:
         result = {}
         for node in nodes:
             try:
-                response = requests.get(
+                response = self.session.get(
                     f"{self._url}{node['id']}/gene/{gene}", verify=self._secure, headers=self._headers
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
@@ -329,7 +338,7 @@ class AgentClient:
         result = {}
         for node in nodes:
             try:
-                response = requests.get(
+                response = self.session.get(
                     f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._secure
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
@@ -358,7 +367,7 @@ class AgentClient:
         result = {}
         for node in nodes:
             try:
-                response = requests.delete(
+                response = self.session.delete(
                     f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._secure
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
@@ -387,7 +396,7 @@ class AgentClient:
         result = {}
         for node in nodes:
             try:
-                response = requests.put(
+                response = self.session.put(
                     f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._secure,
                     json={'model': model}
                 ).json()
@@ -417,7 +426,7 @@ class AgentClient:
         result = {}
         for node in nodes:
             try:
-                response = requests.get(
+                response = self.session.get(
                     f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._secure
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
@@ -436,13 +445,13 @@ class AgentClient:
         """Return name of *nodes*."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.get, 'name', nodes=nodes)
+        return self._query(self.session.get, 'name', nodes=nodes)
 
     def get_time(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Return time on *nodes*."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.get, 'time', nodes=nodes)
+        return self._query(self.session.get, 'time', nodes=nodes)
 
     @_ensure_connected
     def get_vector(self, vector_name: str, nodes: List = None) -> Union[dict, Dict[Any, Dict[str, Any]]]:
@@ -454,7 +463,7 @@ class AgentClient:
         result = {}
         for node in nodes:
             try:
-                response = requests.get(
+                response = self.session.get(
                     f"{self._url}{node['id']}/vector", headers=self._headers, verify=self._secure, json={'data': vector_name}
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
@@ -479,7 +488,7 @@ class AgentClient:
         result = {}
         for node in nodes:
             try:
-                response = requests.post(
+                response = self.session.post(
                     f"{self._url}{node['id']}/gene/increment-recall-threshold",
                     verify=self._secure,
                     headers=self._headers,
@@ -502,19 +511,19 @@ class AgentClient:
         """Tells *nodes* to start sleeping."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.post, 'sleeping/start', nodes=nodes)
+        return self._query(self.session.post, 'sleeping/start', nodes=nodes)
 
     def stop_sleeping(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Wakes up sleeping *nodes*."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.post, 'sleeping/stop', nodes=nodes)
+        return self._query(self.session.post, 'sleeping/stop', nodes=nodes)
 
     def start_predicting(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Tells *nodes* to start predicting."""
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.post, 'predicting/start', nodes=nodes)
+        return self._query(self.session.post, 'predicting/start', nodes=nodes)
 
     def stop_predicting(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Tells *nodes* to stop predicting.
@@ -523,18 +532,18 @@ class AgentClient:
         """
         if nodes is None:
             nodes = self.all_nodes
-        return self._query(requests.post, 'predicting/stop', nodes=nodes)
+        return self._query(self.session.post, 'predicting/stop', nodes=nodes)
 
     @_ensure_connected
     def ping(self, nodes: List = None) -> Union[dict, Any]:
         """Ping a node to ensure it's up."""
         if nodes is None:
-            return requests.get(f'{self._url}gaius-api/ping', headers=self._headers).json()
+            return self.session.get(f'{self._url}gaius-api/ping', headers=self._headers).json()
         else:
             nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
             results = {}
             for node in nodes:
-                response = requests.get(f"{self._url}{node['id']}/ping", verify=self._secure, headers=self._headers).json()
+                response = self.session.get(f"{self._url}{node['id']}/ping", verify=self._secure, headers=self._headers).json()
                 if 'error' in response or response["status"] == 'failed':
                     if len(nodes) == 1 and self.summarize_for_single_node:
                         raise Exception("Request Failure:", response)
@@ -551,9 +560,10 @@ class AgentClient:
         Choose the directory to save in with the directory keyword. Default is in './'."""
         _headers = self._headers
         _headers['Content-Encoding'] = 'gzip'
-        archive = requests.get(
+        archive = self.session.get(
                     f"{self._url}database", 
                     headers=_headers,
+                    json={},
                     verify=self._secure
                 )
         archive_file = join(directory, f'{self.gaius_agent}-{self.name}-{datetime.now()}-kb.archive.gz')
@@ -561,6 +571,88 @@ class AgentClient:
             f.write(archive.content)
         return f"Saved as {archive_file}"
     
+    @_ensure_connected
+    def get_kbs_as_json(self, nodes: List = None, directory='./', separated=False, ids=True):
+        """Returns KBs of specified nodes as JSON Objects
+
+        Args:
+            nodes (List, optional): _description_. Defaults to None.
+            directory (str, optional): _description_. Defaults to './'.
+            separated(bool, optional): store each knowledgebase separately. Defaults to False
+            ids(bool, optional): use primitive_ids as keys for knowledgebase instead of node numbers. Defaults to True
+
+        Returns:
+            str: Descriptive message showning where json files were stored
+        """
+        if nodes is None:
+            nodes = self.all_nodes
+        output = self._query(requests.get, 'get_kb', nodes=nodes)
+        filenames = []
+        if isinstance(output, dict):
+            # print(f'nodes = {nodes}, output = {output.keys()}')
+            altered_output = {}
+            if ids == True:
+                for key in output.keys():
+                    node_id = [node["id"] for node in self.all_nodes if node['name'] == key][0]
+                    altered_output[node_id] = output[key]
+            else:
+                altered_output = output
+            if separated:
+                for key, value in altered_output.items():
+                    archive_file = join(directory, f'{self.gaius_agent}-{self.name}-{key}-{datetime.now()}-kb.json')
+                    filenames.append(archive_file)
+                    with open(archive_file, 'w') as f:
+                        f.write(json.dumps({key: altered_output[key]}))
+            else:
+                archive_file = join(directory, f'{self.gaius_agent}-{self.name}-{datetime.now()}-kb.json')
+                filenames.append(archive_file)
+                with open(archive_file, 'w') as f:
+                    f.write(json.dumps(altered_output))
+                
+        return f"saved kbs as {', '.join(filenames)}"
+    
+    @_ensure_connected
+    def load_kbs_from_json(self, path):
+        """Load KBs from a JSON file
+
+        Args:
+            path (str, required): _description_. path to JSON file where KB is stored.
+            If primitive names (P1, P2, P3, etc.)
+
+        Returns:
+            str: 'success' or 'failed'
+        """
+        
+        # try:
+        node_dict = {}
+        for item in self.all_nodes:
+            node_dict[item['name']] = item['id']
+            node_dict[item['id']] = item['id']
+        
+        with open(path, 'r') as f:
+            kb = json.load(f)
+            for key, value in kb.items():
+                prim_id = key
+                if prim_id in node_dict:
+                    prim_id = node_dict[prim_id]
+                else:
+                    print(f'Warning, node {prim_id} not found in topology, skipping')
+                    continue
+                print(f'loading node {prim_id} from kb for {key}')
+                # print(f'value = {value}')
+                print(f'Url = {self._url}{prim_id}/load_kb')
+                response = self.session.post(f'{self._url}{prim_id}/load_kb', headers={'X-API-KEY' : 'ABCD-1234'}, json={'data': value})
+                # print(response)
+                if response.status_code != 200:
+                    print(f'error loading kb for node {prim_id}: {response.text}')
+                    return 'failed', response.status_code
+                print(f'loading node {prim_id} succeeded')
+        # except Exception as e:
+        #     print(e.args)
+        #     return 'failed'
+        
+        return 'success'
+    
     @_ensure_connected
     def put_kbs(self, archive_file):
         """Uploads KBs from local archive_file file.
@@ -570,7 +662,7 @@ class AgentClient:
         _headers['Content-type'] = 'application/octet-stream'
         with open(archive_file, 'rb') as f:
             data = f.read()
-            response = requests.put(f'{self._url}database', 
+            response = self.session.put(f'{self._url}database', 
                         headers=_headers,
                         verify=self._secure,
                         data=data)
@@ -584,13 +676,17 @@ class AgentClient:
         """
         if nodes is None:
             nodes = self.query_nodes
-        return self._query(requests.post, 'set-target-class', nodes=nodes, data=target_class)
+        return self._query(self.session.post, 'set-target-class', nodes=nodes, data=target_class)
 
     def clear_target_class(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
         """Clears target selection."""
         if nodes is None:
             nodes = self.query_nodes
-        return self._query(requests.post, 'clear-target-class', nodes=nodes)
-
+        return self._query(self.session.post, 'clear-target-class', nodes=nodes)
 
+    @_ensure_connected
+    def tf_attach(self, name, config):
+        return self.session.post(f'{self._url}/thinkflux/{name}/attach', data={'config' : config})
+        
+    
 __all__ = (AgentConnectionError, AgentClient, AgentQueryError)
